"""Module containing tests for the abllib.log module"""

import os
import re

from abllib import log

def test_initialize():
    """Ensure that log initialization works as expected"""

    assert callable(log.initialize)

    log.initialize(log.LogLevel.DEBUG)
    log.add_console_handler()

    logger = log.get_logger("testlogger")

    logger.info("test1234")
    # logging output in console can't be tested cause pylint consumes it first

def test_file_handler():
    """Ensure that file handler works as expected"""

    # cleanup if file is left from previous run
    if os.path.isfile("test.log"):
        os.remove("test.log")

    log.initialize(log.LogLevel.DEBUG)
    log.add_file_handler("test.log")

    logger = log.get_logger()
    logger.debug("the debug message")
    logger.critical("this is critical")

    logger = log.get_logger("submodule")
    logger.info("just an info")
    logger.warning("and a warning")

    logger = log.get_logger("submodule2")
    logger.debug("the final message")

    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()
        assert len(content) == 5
        assert re.match(r"\[.*\] \[DEBUG   \] root: the debug message", content[0])
        assert re.match(r"\[.*\] \[CRITICAL\] root: this is critical", content[1])
        assert re.match(r"\[.*\] \[INFO    \] submodule: just an info", content[2])
        assert re.match(r"\[.*\] \[WARNING \] submodule: and a warning", content[3])
        assert re.match(r"\[.*\] \[DEBUG   \] submodule2: the final message", content[4])

    # cleanup and remove the file handler
    log.initialize()
    os.remove("test.log")

def test_exception_logging():
    """Ensure logging an exception looks beautiful"""

    # cleanup if file is left from previous run
    if os.path.isfile("test.log"):
        os.remove("test.log")

    log.initialize(log.LogLevel.INFO)
    log.add_file_handler("test.log")

    logger = log.get_logger()
    try:
        raise RuntimeError("my message")
    except RuntimeError as err:
        logger.exception(err)

    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()
        assert len(content) == 5
        assert re.match(r"\[.*\] \[ERROR   \] root: my message", content[0])
        assert re.match(r"Traceback \(most recent call last\):", content[1])
        assert re.match(r"  File \".*log_test.py\", line \d+, in test_exception_logging", content[2])
        assert re.match(r"    raise RuntimeError\(\"my message\"\)", content[3])
        assert re.match(r"RuntimeError: my message", content[4])

    # cleanup and remove the file handler
    log.initialize()
    os.remove("test.log")

def test_file_handler_delay():
    """Ensure that file handler only creates logfile when needed"""

    # cleanup if file is left from previous run
    if os.path.isfile("test.log"):
        os.remove("test.log")

    log.initialize(log.LogLevel.WARNING)
    assert not os.path.isfile("test.log")

    log.add_file_handler("test.log")
    assert not os.path.isfile("test.log")

    logger = log.get_logger()
    assert not os.path.isfile("test.log")

    logger.debug("debugmsg")
    assert not os.path.isfile("test.log")

    logger.info("infomsg")
    assert not os.path.isfile("test.log")

    logger.warning("warnmsg")
    assert os.path.isfile("test.log")

    with open("test.log", "r", encoding="utf8") as f:
        content = f.read()
        assert content != ""
        assert re.match(r"\[.*\] \[WARNING \] root: warnmsg\n", content)

    # cleanup and remove the file handler
    log.initialize()
    os.remove("test.log")

def test_deprecated_levels():
    """Ensure that LogLevel.WARN and LogLevel.FATAL cannot be used"""

    assert hasattr(log.LogLevel, "WARNING")
    assert not hasattr(log.LogLevel, "WARN")

    assert hasattr(log.LogLevel, "CRITICAL")
    assert not hasattr(log.LogLevel, "FATAL")

def test_initialize_invalidtypes():
    """Ensure that initialize only accepts valid arguments"""

    # to be added when abllib.type module is implemented



"""The main module, used for dev testing"""

# TODO:
# UP NEXT
# replace all TypeError occurances with WrongTypeError

# IMPORTANT BUT ANNOYING
# add tests for threaded storage access
# document levenshtein_distance
# document fuzzy score calculation
# optimize _lock_wrapper._log_callstack (takes half of the runtime when assigning storage item)

# TAKES A LOT OF TIME / NEEDED IN THE FUTURE
# construct target type-specific function at runtime in types.enforce 

# pylint: skip-file

if __name__ == "__main__":
    pass



"""Module containing the log_io wrapper"""

import functools
from typing import Callable, Any

from ._base_log_wrapper import BaseLogWrapper

class log_io(BaseLogWrapper):
    """
    Decorate a function, which logs any passed arguments and return values.
    The values are logged with log level DEBUG, so make sure you configured your logger properly.

    If the optional argument logger is set and of type logging.Logger, log the values to that logger.

    If the optional argument logger is set and of type str, request that logger and log the values.

    Otherwise, the values are logged to the root logger.
    """

    def __call__(self, func: Callable):
        """Called when the class instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            res = func(*args, **kwargs)

            self.logger.debug(f"func: {func.__name__}")
            form_args = [self._format(item) for item in args]
            form_kwargs = [f"{key}={self._format(value)}" for key, value in kwargs.items()]
            arg_str = ", ".join(form_args + form_kwargs)
            self.logger.debug(f"in  : {arg_str}")
            self.logger.debug(f"out : {self._format(res)}")

            return res

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper

    def _format(self, arg: Any) -> str:
        if isinstance(arg, str):
            return f"\"{arg}\""

        return str(arg)



"""A module containing file name-modification functions."""

import sys
from typing import Generator

try:
    # optional module for japanese character transliterating
    import pykakasi
except ImportError:
    pass

from abllib.error import WrongTypeError

CHARS_TO_REMOVE = "',#^?!\"<>%$%°*"
CHARS_TO_REPLACE = " /\\|~+:;@\n"

def sanitize(filename: str) -> str:
    """
    Return a sanitized version of the file name, which replaces/removes all invalid symbols.

    Additionally, all spaces are replaced with underscores.
    """

    if not isinstance(filename, str):
        raise WrongTypeError.with_values(filename, str)

    filename = filename.strip(" ")

    # remove trailing chars
    filename = filename.strip(CHARS_TO_REPLACE)

    filename = _sanitize_letters(filename)

    filename = _sanitize_punctuations(filename)

    filename = _sanitize_symbols(filename)

    return filename

def _sanitize_letters(filename: str) -> str:
    """convert invalid letters"""

    # german Umlaute
    # https://en.wikipedia.org/wiki/Umlaut_(diacritic)
    filename = filename.replace("ä", "a")
    filename = filename.replace("Ä", "A")
    filename = filename.replace("ö", "o")
    filename = filename.replace("Ö", "O")
    filename = filename.replace("ü", "u")
    filename = filename.replace("Ü", "U")
    filename = filename.replace("ß", "ss")

    # japanese characters
    if _contains_japanese_char(filename):
        if "pykakasi" in sys.modules:
            filename = _replace_japanese_chars(filename)
        else:
            # needs to be imported here to prevent circular import
            # pylint: disable-next=cyclic-import, import-outside-toplevel
            from abllib import log
            logger = log.get_logger("sanitize")
            logger.warning("to properly transliterate japanese text to rōmaji, "
                           "you need to install the optional dependency 'pykakasi'")

    return filename

def _sanitize_punctuations(filename: str) -> str:
    """make punctuation marks readable"""

    # to make sentences seem reasonable
    filename = filename.replace(", ", "_")
    filename = filename.replace(". ", "_")
    filename = filename.replace("! ", "_")
    filename = filename.replace("? ", "_")
    filename = filename.replace("; ", "_")
    filename = filename.replace(": ", "_")
    filename = filename.replace(" \n", "_")

    # fix sentences ending in a dot
    filename = filename.replace("..", ".")

    return filename

def _sanitize_symbols(filename: str) -> str:
    """remove/replace invalid symbols"""

    for char in CHARS_TO_REMOVE:
        filename = filename.replace(char, "")

    for char in CHARS_TO_REPLACE:
        filename = filename.replace(char, "_")

    # remove leftover non-ascii characters
    filename = filename.encode('ascii', 'ignore').decode()

    return filename

# original code from here:
# https://stackoverflow.com/a/30070664/15436169
japanese_char_ranges = [
    {"from": ord("\u3300"), "to": ord("\u33ff")},         # compatibility ideographs
    {"from": ord("\ufe30"), "to": ord("\ufe4f")},         # compatibility ideographs
    {"from": ord("\uf900"), "to": ord("\ufaff")},         # compatibility ideographs
    {"from": ord("\U0002F800"), "to": ord("\U0002fa1f")}, # compatibility ideographs
    {"from": ord("\u3040"), "to": ord("\u309f")},         # Japanese Hiragana
    {"from": ord("\u30a0"), "to": ord("\u30ff")},         # Japanese Katakana
    {"from": ord("\u2e80"), "to": ord("\u2eff")},         # cjk radicals supplement
    {"from": ord("\u4e00"), "to": ord("\u9fff")},
    {"from": ord("\u3400"), "to": ord("\u4dbf")},
    {"from": ord("\U00020000"), "to": ord("\U0002a6df")},
    {"from": ord("\U0002a700"), "to": ord("\U0002b73f")},
    {"from": ord("\U0002b740"), "to": ord("\U0002b81f")},
    {"from": ord("\U0002b820"), "to": ord("\U0002ceaf")}  # included as of Unicode 8.0
]

def _contains_japanese_char(text) -> bool:
    for char in text:
        if _is_japanese_letter(char):
            return True

    return False

def _is_japanese_letter(char: str) -> bool:
    if char == " ":
        return False

    for char_range in japanese_char_ranges:
        if char_range["from"] <= ord(char) <= char_range["to"]:
            return True

    return False

def _replace_japanese_chars(text: str) -> Generator[str, None, None]:
    # replace japanese Full stop (https://en.wikipedia.org/wiki/Japanese_punctuation#Full_stop)
    text = text.replace("。", ". ")

    i = 0
    while i < len(text):
        if _is_japanese_letter(text[i]):
            start = i

            # the current character is known to be japanese
            i += 1

            while i < len(text) and _is_japanese_letter(text[i]):
                i += 1

            converted_text = " ".join([item["hepburn"] for item in pykakasi.kakasi().convert(text[start:i])])

            text = text[:start] + converted_text + text[i:]
        i += 1

    return text



"""
Ableytner's library for Python

Contains many general-purpose functions which can be used across projects.
"""

from . import alg, error, fs, fuzzy, log, onexit, pproc, storage, wrapper
from .log import get_logger, LogLevel
from .storage import CacheStorage, VolatileStorage, PersistentStorage, StorageView
from .wrapper import Lock, Semaphore, NamedLock, NamedSemaphore

__exports__ = [
    alg,
    error,
    fs,
    fuzzy,
    log,
    onexit,
    pproc,
    storage,
    wrapper,
    get_logger,
    LogLevel,
    Lock,
    Semaphore,
    NamedLock,
    NamedSemaphore,
    CacheStorage,
    VolatileStorage,
    PersistentStorage,
    StorageView
]



"""Module containing the pylint test"""

import os

import pytest

from abllib import fs, log

logger = log.get_logger("pylint")

def test_pylint():
    """Checks if all git-tracked python files adhere to pylint rules"""

    ROOTDIR = fs.absolute(__file__, "..", "..", "..")
    PREV_DIR = os.getcwd()
    os.chdir(ROOTDIR)

    if os.name == "nt":
        files = os.popen("git ls-files *.py").read()
        files = " ".join([file.strip() for file in files.split("\n")])

        pylint_output = []
        for line in os.popen(f"python -m pylint {files}").readlines():
            if line.strip().strip("-") != "":
                pylint_output.append(line.strip())

        os.chdir(PREV_DIR)

        if len(pylint_output) == 0:
            pytest.fail("Detected error during test. Is pylint installed?")

        if not "Your code has been rated at 10.00/10" in pylint_output[-1]:
            for line in pylint_output:
                logger.warning(line)

            pytest.fail(pylint_output[-1])
    else:
        # logging pylint errors on linux doesn't work
        pylint_output = os.popen("python3 -m pylint $(git ls-files '*.py')").read()

        os.chdir(PREV_DIR)

        assert "Your code has been rated at 10.00/10" in pylint_output



"""A module containing various wrappers"""

from ._deprecated import deprecated
from ._lock import Lock, Semaphore
from ._lock_wrapper import NamedLock, NamedSemaphore, ReadLock, WriteLock
from ._log_error import log_error
from ._log_io import log_io
from ._singleuse_wrapper import singleuse
from ._timeit import timeit

__exports__ = [
    Lock,
    NamedLock,
    NamedSemaphore,
    ReadLock,
    Semaphore,
    WriteLock,
    deprecated,
    log_error,
    log_io,
    timeit,
    singleuse
]



"""Module containing tests for the abllib.wrapper module"""

# pylint: disable=function-redefined, consider-using-with, unused-argument

import re
import os
from datetime import datetime
from time import sleep

import pytest

from abllib import error, wrapper, log
from abllib.pproc import WorkerThread

def test_lock():
    """Ensure that Lock works as expected"""

    assert hasattr(wrapper, "Lock")
    assert callable(wrapper.Lock)

    lock = wrapper.Lock()

    assert not lock.locked()
    assert lock.acquire(blocking=True, timeout=1)
    assert lock.locked()

    assert not lock.acquire(blocking=True, timeout=1)

    lock.release()
    assert not lock.locked()

def test_semaphore():
    """Ensure that Semaphore works as expected"""

    assert hasattr(wrapper, "Semaphore")
    assert callable(wrapper.Semaphore)

    sem = wrapper.Semaphore(3)

    assert not sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()
    assert sem.acquire(blocking=True, timeout=1)
    assert sem.locked()

    # the semaphore is full
    assert not sem.acquire(blocking=True, timeout=1)

    sem.release()
    assert sem.locked()
    sem.release()
    assert sem.locked()
    sem.release()
    assert not sem.locked()

def test_namedlock():
    """Ensure that NamedLock works as expected"""

    assert hasattr(wrapper, "NamedLock")
    assert callable(wrapper.NamedLock)

    @wrapper.NamedLock("test1", timeout=0.1)
    def func1():
        return True

    assert not wrapper.NamedLock("test1").locked()
    assert func1()
    assert not wrapper.NamedLock("test1").locked()

    wrapper.NamedLock("test2").acquire()
    assert wrapper.NamedLock("test2").locked()
    assert wrapper.NamedLock("test2", timeout=1).locked()

    wrapper.NamedLock("test3").acquire()
    def func2():
        assert wrapper.NamedLock("test3").locked()
        wrapper.NamedLock("test3", timeout=4).acquire()

    start_time = datetime.now()
    thread = WorkerThread(target=func2)
    thread.start()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        thread.join(reraise=True)

    duration = datetime.now() - start_time
    assert duration.total_seconds() > 3.5
    assert duration.total_seconds() < 4.5

def test_namedsemaphore():
    """Ensure that NamedSemaphore works as expected"""

    assert hasattr(wrapper, "NamedSemaphore")
    assert callable(wrapper.NamedSemaphore)

    @wrapper.NamedSemaphore("test1", timeout=0.1)
    def func1():
        return True

    assert not wrapper.NamedSemaphore("test1").locked()
    assert func1()
    assert not wrapper.NamedSemaphore("test1").locked()

    wrapper.NamedSemaphore("test2").acquire()
    assert wrapper.NamedSemaphore("test2").locked()
    assert wrapper.NamedSemaphore("test2", timeout=1).locked()

    wrapper.NamedLock("test3").acquire()
    def func2():
        assert not wrapper.NamedSemaphore("test3").locked()
        wrapper.NamedSemaphore("test3", timeout=4).acquire()

    start_time = datetime.now()
    thread = WorkerThread(target=func2)
    thread.start()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        thread.join(reraise=True)

    duration = datetime.now() - start_time
    assert duration.total_seconds() > 3.5
    assert duration.total_seconds() < 4.5

def test_namedlocks_combined():
    """Ensure that NamedLock and NamedSemaphore work together correctly"""

    @wrapper.NamedLock("test1", timeout=0.1)
    def func():
        return True

    wrapper.NamedSemaphore("test1").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test1").release()


    @wrapper.NamedLock("test2", timeout=0.1)
    def func():
        return True

    wrapper.NamedSemaphore("test2").acquire()
    wrapper.NamedSemaphore("test2").acquire()
    wrapper.NamedSemaphore("test2").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedSemaphore("test2").release()
    func()


    @wrapper.NamedSemaphore("test3", timeout=0.1)
    def func():
        return True

    wrapper.NamedLock("test3").acquire()
    with pytest.raises(error.LockAcquisitionTimeoutError):
        func()
    wrapper.NamedLock("test3").release()

def test_locks_underscore_names():
    """Ensure that named lock names can start with an underscore"""

    lock = wrapper.NamedSemaphore("_test1")
    assert not lock.locked()
    lock.acquire()
    assert lock.locked()
    assert wrapper.NamedSemaphore("_test1").locked()
    lock.release()
    assert not lock.locked()
    assert not wrapper.NamedSemaphore("_test1").locked()

    lock = wrapper.NamedLock("_test2")
    assert not lock.locked()
    lock.acquire()
    assert lock.locked()
    assert wrapper.NamedLock("_test2").locked()
    lock.release()
    assert not lock.locked()
    assert not wrapper.NamedLock("_test2").locked()

def test_singleuse():
    """Ensure that singleuse works as expected"""

    @wrapper.singleuse
    def func1():
        pass

    func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

def test_singleuse_exception():
    """Ensure that raised exceptions reset singleuse flag"""

    data = [1, 2, 3]

    @wrapper.singleuse
    def func1():
        if len(data) > 0:
            data.pop(0)
            raise error.InternalCalculationError()

    with pytest.raises(error.InternalCalculationError):
        func1()
    with pytest.raises(error.InternalCalculationError):
        func1()
    with pytest.raises(error.InternalCalculationError):
        func1()

    func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

    with pytest.raises(error.CalledMultipleTimesError):
        func1()

def test_log_error_default(capture_logs):
    """Ensure that log_error uses the root logger by default"""

    @wrapper.log_error
    def func1():
        raise RuntimeError("my message")

    with pytest.raises(RuntimeError):
        func1()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        # remove "pointer" lines only present in python 3.12
        content = list(filter(lambda x: x.strip().strip("^") != "", content))

        assert len(content) == 7
        assert re.match(r"\[.*\] \[ERROR   \] root: my message", content[0])
        assert re.match(r"Traceback \(most recent call last\):", content[1])
        assert re.match(r"  File \".*_log_error.py\", line \d+, in wrapper", content[2])
        assert re.match(r"    return func\(\*args, \*\*kwargs\)", content[3])
        assert re.match(r"  File \".*wrapper_test.py\", line \d+, in func1", content[4])
        assert re.match(r"    raise RuntimeError\(\"my message\"\)", content[5])
        assert re.match(r"RuntimeError: my message", content[6])

def test_log_error_loggername(capture_logs):
    """Ensure that log_error uses the provided logger name"""

    @wrapper.log_error("mycustomlogger")
    def func1():
        raise RuntimeError("my message")

    with pytest.raises(RuntimeError):
        func1()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        # remove "pointer" lines only present in python 3.12
        content = list(filter(lambda x: x.strip().strip("^") != "", content))

        assert len(content) == 7
        assert re.match(r"\[.*\] \[ERROR   \] mycustomlogger: my message", content[0])
        assert re.match(r"Traceback \(most recent call last\):", content[1])
        assert re.match(r"  File \".*_log_error.py\", line \d+, in wrapper", content[2])
        assert re.match(r"    return func\(\*args, \*\*kwargs\)", content[3])
        assert re.match(r"  File \".*wrapper_test.py\", line \d+, in func1", content[4])
        assert re.match(r"    raise RuntimeError\(\"my message\"\)", content[5])
        assert re.match(r"RuntimeError: my message", content[6])

def test_log_error_customlogger(capture_logs):
    """Ensure that log_error uses the provided custom logger"""

    @wrapper.log_error(log.get_logger("mycustomlogger"))
    def func1():
        raise RuntimeError("my message")

    with pytest.raises(RuntimeError):
        func1()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        # remove "pointer" lines only present in python 3.12
        content = list(filter(lambda x: x.strip().strip("^") != "", content))

        assert len(content) == 7
        assert re.match(r"\[.*\] \[ERROR   \] mycustomlogger: my message", content[0])
        assert re.match(r"Traceback \(most recent call last\):", content[1])
        assert re.match(r"  File \".*_log_error.py\", line \d+, in wrapper", content[2])
        assert re.match(r"    return func\(\*args, \*\*kwargs\)", content[3])
        assert re.match(r"  File \".*wrapper_test.py\", line \d+, in func1", content[4])
        assert re.match(r"    raise RuntimeError\(\"my message\"\)", content[5])
        assert re.match(r"RuntimeError: my message", content[6])

def test_log_error_handler():
    """Ensure that log_error uses the provided handler"""

    results = []
    def myhandler(exc_text):
        results.append(exc_text)

    @wrapper.log_error(handler=myhandler)
    def func1():
        raise RuntimeError("my message")

    with pytest.raises(RuntimeError):
        func1()

    assert len(results) == 1
    assert results[0] == "RuntimeError: my message"

def test_log_io_default(capture_logs):
    """Ensure that log_io uses the root logger by default"""

    @wrapper.log_io
    def func1(name, age, message = ""):
        return f"{name}: {message}"

    assert func1("Tom", 46) == "Tom: "
    assert func1("Anna", 77, message="no") == "Anna: no"

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 6
        assert re.match(r'\[.*\] \[DEBUG   \] root: func: func1', content[0])
        assert re.match(r'\[.*\] \[DEBUG   \] root: in  : "Tom", 46', content[1])
        assert re.match(r'\[.*\] \[DEBUG   \] root: out : "Tom: "', content[2])
        assert re.match(r'\[.*\] \[DEBUG   \] root: func: func1', content[3])
        assert re.match(r'\[.*\] \[DEBUG   \] root: in  : "Anna", 77, message="no"', content[4])
        assert re.match(r'\[.*\] \[DEBUG   \] root: out : "Anna: no"', content[5])

def test_log_io_loggername(capture_logs):
    """Ensure that log_io uses the provided logger name"""

    @wrapper.log_io("SpecialLogger")
    def func1(name, age, message = ""):
        return f"{name}: {message}"

    assert func1("Tom", 46) == "Tom: "
    assert func1("Anna", 77, "no") == "Anna: no"

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 6
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: func: func1', content[0])
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: in  : "Tom", 46', content[1])
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: out : "Tom: "', content[2])
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: func: func1', content[3])
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: in  : "Anna", 77, "no"', content[4])
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: out : "Anna: no"', content[5])

def test_log_io_custom_logger(capture_logs):
    """Ensure that log_io uses a custom provided logger"""

    @wrapper.log_io(log.get_logger("ExtraLogger"))
    def func1(name, age, message = ""):
        return f"{name}: {message}"

    assert func1("Tom", 46) == "Tom: "
    assert func1("Anna", 77, "no") == "Anna: no"

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 6
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: func: func1', content[0])
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: in  : "Tom", 46', content[1])
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: out : "Tom: "', content[2])
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: func: func1', content[3])
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: in  : "Anna", 77, "no"', content[4])
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: out : "Anna: no"', content[5])

def test_timeit_default(capture_logs):
    """Ensure that timeit uses the root logger by default"""

    @wrapper.timeit
    def func1():
        sleep(0.001)
    @wrapper.timeit
    def func2():
        sleep(0.01)
    @wrapper.timeit
    def func3():
        sleep(0.1)
    @wrapper.timeit
    def func4():
        sleep(1)
    @wrapper.timeit
    def func5():
        sleep(10)


    func1()
    func2()
    func3()
    func4()
    func5()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 5
        assert re.match(r'\[.*\] \[DEBUG   \] root: func1: \d{1}\.\d{2} ms elapsed', content[0])
        assert re.match(r'\[.*\] \[DEBUG   \] root: func2: \d{2}\.\d{2} ms elapsed', content[1])
        assert re.match(r'\[.*\] \[DEBUG   \] root: func3: \d{3}\.\d{2} ms elapsed', content[2])
        assert re.match(r'\[.*\] \[DEBUG   \] root: func4: \d{1}\.\d{2} s elapsed', content[3])
        assert re.match(r'\[.*\] \[DEBUG   \] root: func5: \d{2}\.\d{2} s elapsed', content[4])

def test_timeit_loggername(capture_logs):
    """Ensure that timeit uses the provided logger name"""

    @wrapper.timeit("SpecialLogger")
    def func1():
        sleep(0.001)

    func1()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 1
        assert re.match(r'\[.*\] \[DEBUG   \] SpecialLogger: func1: \d{1}\.\d{2} ms elapsed', content[0])

def test_timeit_custom_logger(capture_logs):
    """Ensure that timeit uses a custom provided logger"""

    @wrapper.timeit(log.get_logger("ExtraLogger"))
    def func1():
        sleep(0.001)

    func1()

    assert os.path.isfile("test.log")
    with open("test.log", "r", encoding="utf8") as f:
        content = f.readlines()

        assert len(content) == 1
        assert re.match(r'\[.*\] \[DEBUG   \] ExtraLogger: func1: \d{1}\.\d{2} ms elapsed', content[0])



"""Module containing the timeit wrapper"""

import functools
from time import perf_counter_ns
from typing import Callable

from ._base_log_wrapper import BaseLogWrapper

class timeit(BaseLogWrapper):
    """
    Decorate a function, which logs the execution time of this function.
    The values are logged with log level DEBUG, so make soure you configured your logger properly.

    If the optional argument logger is set and of type logging.Logger, log the time to that logger.

    If the optional argument logger is set and of type str, request that logger and log the time.

    Otherwise, the time is logged to the root logger.
    """

    def __call__(self, func: Callable):
        """Called when the class instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            start = perf_counter_ns()

            res = func(*args, **kwargs)

            elapsed = perf_counter_ns() - start

            log_msg = f"{func.__name__}: "

            for unit in ["ns", "μs", "ms"]:
                if elapsed < 1000:
                    log_msg += f"{elapsed:3.2f} {unit} elapsed"
                    self.logger.debug(log_msg)
                    return res

                elapsed /= 1000

            for unit in ["s", "min"]:
                if elapsed < 60:
                    log_msg += f"{elapsed:2.2f} {unit} elapsed"
                    self.logger.debug(log_msg)
                    return res

                elapsed /= 60

            if elapsed < 24:
                log_msg += f"{elapsed:2.2f} hours elapsed"
                self.logger.debug(log_msg)
                return res

            elapsed /= 24
            log_msg += f"{elapsed:.2f} days elapsed"
            self.logger.debug(log_msg)
            return res

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper



"""Module containing tests for the abllib.fs module"""

# pylint: disable=missing-class-docstring

import pathlib
import sys
import os

import pytest

from abllib import fs

def test_absolute():
    """Ensure that fs.absolute works as expected"""

    assert callable(fs.absolute)
    assert fs.absolute(os.getcwd()) == os.getcwd().replace("c:\\", "C:\\")
    assert fs.absolute("test.txt") == os.path.join(_uppercase_path(os.getcwd()), "test.txt")
    assert fs.absolute("subdir", "another", "test.txt") \
           == os.path.join(_uppercase_path(os.getcwd()), "subdir", "another", "test.txt")
    assert fs.absolute("subdir", pathlib.Path("another"), "test.txt") \
           == os.path.join(_uppercase_path(os.getcwd()), "subdir", "another", "test.txt")
    assert fs.absolute("subdir", "..", "test.txt") == os.path.join(_uppercase_path(os.getcwd()), "test.txt")

    with pytest.raises(TypeError):
        fs.absolute(None)
    with pytest.raises(TypeError):
        fs.absolute(1)
    with pytest.raises(TypeError):
        fs.absolute("one", "two", 3)
    with pytest.raises(ValueError):
        fs.absolute()

def _uppercase_path(path: str) -> str:
    if path.startswith("/"):
        return path

    first = path[0]
    first = first.upper()
    return first + path[1:]

def test_sanitize():
    """Ensure that fs.sanitize works as expected"""

    assert callable(fs.sanitize)
    assert fs.sanitize("title") == "title"
    assert fs.sanitize("myimage.jpeg") == "myimage.jpeg"
    assert fs.sanitize("myfilename.txt") == "myfilename.txt"

    # spaces
    assert fs.sanitize("this is a normal sentence") == "this_is_a_normal_sentence"
    assert fs.sanitize("the sentence ends.") == "the_sentence_ends."

    # punctuation marks
    assert fs.sanitize("This sentence gets converted..txt") == "This_sentence_gets_converted.txt"
    assert fs.sanitize("the sentence ends..txt") == "the_sentence_ends.txt"
    assert fs.sanitize("sure?") == "sure"
    assert fs.sanitize("sure!") == "sure"
    assert fs.sanitize("I'm sure. Are you?") == "Im_sure_Are_you"

    # long sentences
    # pylint: disable-next=line-too-long
    assert fs.sanitize("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vivamus venenatis aliquet turpis, ac laoreet mauris fermentum eu. Suspendisse luctus a ante sit amet vehicula. Vestibulum aliquet non dui at efficitur. Donec fermentum nulla sit amet eros viverra accumsan. Sed et tortor sit amet diam tempor condimentum sed ut erat. Curabitur cursus dignissim tincidunt.") == "Lorem_ipsum_dolor_sit_amet_consectetur_adipiscing_elit_Vivamus_venenatis_aliquet_turpis_ac_laoreet_mauris_fermentum_eu_Suspendisse_luctus_a_ante_sit_amet_vehicula_Vestibulum_aliquet_non_dui_at_efficitur_Donec_fermentum_nulla_sit_amet_eros_viverra_accumsan_Sed_et_tortor_sit_amet_diam_tempor_condimentum_sed_ut_erat_Curabitur_cursus_dignissim_tincidunt."

    # newlines
    assert fs.sanitize("first line\nnext line") == "first_line_next_line"
    assert fs.sanitize("first line \nnext line") == "first_line_next_line"

    # special characters
    assert fs.sanitize("special' char/act\\ers ar|e i*gnor;ed") == "special_char_act_ers_ar_e_ignor_ed"

    # german Umlaute
    assert fs.sanitize("Äpfel") == "Apfel"
    assert fs.sanitize("Die grüne Böschung") == "Die_grune_Boschung"
    assert fs.sanitize("Straße") == "Strasse"

    # japanese characters
    assert fs.sanitize("ハウルの動く城") == "hauru_no_ugoku_shiro"
    assert fs.sanitize("葬送のフリーレン      ") == "sousou_no_furiiren"
    assert fs.sanitize("こんにちは World!") == "konnichiha_World"
    assert fs.sanitize("Hello 世界!") == "Hello_sekai"

    # pykakasi is optional
    mod = sys.modules.pop("pykakasi")
    assert fs.sanitize("ハウルの動く城") == ""
    assert fs.sanitize("葬送のフリーレン      ") == ""
    assert fs.sanitize("こんにちは World!") == "_World"
    assert fs.sanitize("Hello 世界!") == "Hello_"
    sys.modules["pykakasi"] = mod

    # mixed sentences
    assert fs.sanitize("最初の文。The second sentence.") == "saisho_no_bun_The_second_sentence."
    assert fs.sanitize("最初の文。The second sentence..txt") == "saisho_no_bun_The_second_sentence.txt"



"""Module containing the log_error wrapper"""

import functools
from logging import Logger
from typing import Callable

from abllib.error import ArgumentCombinationError
from ._base_log_wrapper import BaseLogWrapper

class log_error(BaseLogWrapper):
    """
    Decorate a function, which logs any exception occuring during execution.

    If the optional argument logger is set and of type logging.Logger, log the error to that logger.

    If the optional argument logger is set and of type str, request that logger and log the error.

    If the optional argument handler is set, forwards the error message to that function.

    Otherwise, the error is logged to the root logger.
    """

    def __new__(cls, logger: str | Logger | None | Callable = None, handler: Callable | None = None):
        if logger is None and handler is None:
            raise ArgumentCombinationError("Either logger or handler need to be provided")

        # handler is given
        if handler is not None:
            inst = super().__new__(cls)
            # pylint: disable-next=unnecessary-lambda-assignment
            inst.logger = handler
            return inst

        return super().__new__(cls, logger)

    def __call__(self, func: Callable):
        """Called when the class instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            try:
                return func(*args, **kwargs)
            except Exception as e:
                if isinstance(self.logger, Logger):
                    self.logger.exception(e)
                else:
                    self.logger(f"{e.__class__.__name__}: {e}")
                raise

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper



"""Module containing the log_io wrapper"""

import functools
from logging import Logger
from typing import Callable

from abllib import log
from abllib.error import WrongTypeError

class BaseLogWrapper():
    """
    Base class for all decorators that log something.

    If the optional argument logger is set and of type logging.Logger, use that logger.

    If the optional argument logger is set and of type str, request and use that loggers.

    Otherwise, use the root logger.
    """

    def __new__(cls, logger: str | Logger | Callable | None = None):
        inst = super().__new__(cls)

        # used directly as a wrapper
        if callable(logger):
            inst.logger = log.get_logger()
            return inst(logger)

        # logger is the loggers' name
        if isinstance(logger, str):
            _logger = log.get_logger(logger)
        # logger is a logging.Logger object
        elif isinstance(logger, Logger):
            _logger = logger
        elif logger is None:
            _logger = None
        else:
            raise WrongTypeError.with_values(logger, str, Logger, Callable, None)

        inst.logger = _logger
        return inst

    logger: Logger | None

    def __call__(self, func: Callable):
        """Called when the class instance is used as a decorator"""

        def wrapper(*args, **kwargs):
            """The wrapped function that is called on function execution"""

            raise NotImplementedError()

        # https://stackoverflow.com/a/17705456/15436169
        functools.update_wrapper(wrapper, func)

        return wrapper
